"""
Modulo Commons del ecosistema Kloop. Contiene los modulos de uso común para los paquetes 
(1) categorization
(2) reporting

PyPI URL:
"""
import sys
import logging

if sys.version_info[:2] >= (3, 10):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.10`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.2"

    log = logging.getLogger(__name__)


class ControllerMLFlow:
    def __init__(self):
        pass
        # self.command_dir_logs = 'mkdir ~/mlflow/mllogs'
        # self.command_dir_experiments = 'mkdir ~/mlflow/mlruns'

    def create_dir():
        pass

    def run_mlflow(self):
        pass



from os import getenv
from pykafka import KafkaClient
from pykafka.common import OffsetType
import math
from itertools import islice
import json

null=None; true=True; false=False

from kafka import KafkaProducer
from json import dumps as dumps_


class Producer:
    def __init__(
                self, 
                topic = None,
                kafka_brokers = None,
                acks = None,
                api_version = None,
                client_id = None
                 ):
        self.kafka_api_version = api_version
        self.topic = topic
        self.acks = acks
        self.client_id = client_id
        self.kafka_brokers = kafka_brokers
        

    def init_producer(self) -> None:

        self.producer = KafkaProducer(
            acks = self.acks,
            api_version=self.kafka_api_version,
            bootstrap_servers = self.kafka_brokers,
            value_serializer = lambda x: dumps_(x).encode('utf-8'),
        )

    def send_message(self, data: dict) -> None:
        self.producer.send(self.topic, value=data)
        # block until all async messages are sent
        self.producer.flush()


            
def get_offset_latest(topics,kafka_brokers):
    
    latest_offsets = {k: 0 for k in topics}
    earliest_available_offsets = {k: 0 for k in topics}
    
    
    for topic_ in topics:
        client = KafkaClient(hosts=kafka_brokers)
        topic = client.topics[topic_]

        consumer = topic.get_simple_consumer(auto_offset_reset=OffsetType.LATEST,reset_offset_on_start=True)

        earliest_available_offsets[topic_] = topic.earliest_available_offsets()[0][0][0]
        latest_offsets[topic_] = topic.latest_available_offsets()[0][0][0]
    
    return earliest_available_offsets,latest_offsets





def get_send_messages(topic_,last_num_messages,kafka_brokers):
    list_type_messages = []
    list_content_messages = []
    client = KafkaClient(hosts=kafka_brokers)
    
    topic = client.topics[topic_]
    

    consumer = topic.get_simple_consumer(auto_offset_reset=OffsetType.LATEST,reset_offset_on_start=True)


    # how many messages should we get from the end of each partition?
    MAX_PARTITION_REWIND = int(math.ceil(last_num_messages / len(consumer._partitions)))

    # find the beginning of the range we care about for each partition
    offsets = [(p, op.last_offset_consumed - MAX_PARTITION_REWIND) for p, op in consumer._partitions.items()]

    print(offsets)

    # if we want to rewind before the beginning of the partition, limit to beginning
    offsets = [(p, (o if o > -1 else -2)) for p, o in offsets]

    # reset the consumer's offsets
    consumer.reset_offsets(offsets)

    message_count = 0
    message_categorized_count = 0
    for message in islice(consumer, last_num_messages):
        print("offset: ",message.offset)
        current_offset = message.offset
        message = json.loads(message.value)
        if message['type'] ==  'transaction-categorized':
            message_categorized_count += 1
            for transaction in message['data']['transactions']:
                list_content_messages.append({'uid':transaction['uid'],'kw_group':transaction['categorization']['kw_group']})

        message_count += 1
        list_type_messages.append(message["type"])
        
        if message_count == last_num_messages :
            break

    print("message_count: ", message_count)
    print("message_categorized_count: ", message_categorized_count) 

        
    return list_type_messages,list_content_messages



if __name__ == '__main__':
    
    confluence_library= False
    port_current = '9094'
    port_new = '9092'
    default_kafka_host = 'localhost:9092'
    enable_port_replace = True


    if getenv('KAFKA_BROKERS'):
        if len(getenv('KAFKA_BROKERS')) != 0:
            kafka_brokers = getenv('KAFKA_BROKERS')


    elif getenv('KAFKA_HOST') and getenv('KAFKA_PORT'):
        kafka_brokers = getenv('KAFKA_HOST') + ':' + getenv('KAFKA_PORT')
    else :
        kafka_brokers = default_kafka_host

    kafka_brokers = kafka_brokers.replace(port_current,port_new)

    print("kafka_brokers: ", kafka_brokers)
    
   



    acks = 'all'
    api_version = (2,6,1)
    group_id  = 'categorization-'
    topic_set = 'transaction'

    topics_get = ['transaction']
    message_type = 'transaction-update-kw-group'
    earliest_available_offsets,latest_offsets = get_offset_latest(topics_get,kafka_brokers)
    message_struct = {'data':{
                            'transactions':None
                            },
                   'type': 'transaction-update-kw-group'
                  }




    print(earliest_available_offsets )
    print(latest_offsets)

    
    for topic in topics_get :
        print("topic: ", topic)
        last_num_messages = latest_offsets[topic] - earliest_available_offsets[topic]
        print("last_num_messages: ", last_num_messages)
        _,list_content_messages = get_send_messages(topic,last_num_messages,kafka_brokers)

        
        print("Número de transacciones: ",len(list_content_messages))

        # d_unique = pd.DataFrame(d).drop_duplicates().to_dict('records')
        set_of_jsons = {json.dumps(d, sort_keys=True) for d in list_content_messages}
        list_content_messages = [json.loads(t) for t in set_of_jsons]
        
        print("Remover duplicados... \nNúmero de transacciones: ",len(list_content_messages))
        
        message_struct['data']['transactions'] = list_content_messages
        
        if confluence_library :
            if enable_port_replace:
                kafka_brokers = kafka_brokers.replace(port_current,port_new)
        else :
            if enable_port_replace:
                kafka_brokers = kafka_brokers.replace(port_current,port_new).split(',')


        print("kafka_brokers: ", kafka_brokers)
        
        acks = 'all'
        api_version = (2,6,1)
        group_id  = 'categorization-update-kw-group'
        client_id = None
        ProdTrans = Producer('transaction', kafka_brokers, acks, api_version , client_id)
        ProdTrans.init_producer()
        ProdTrans.send_message(message_struct)
        print(message_struct)



from logging import basicConfig
from logging import getLogger
from logging import INFO
from logging import WARNING
from logging import ERROR
from logging import CRITICAL
from logging import DEBUG
from os import getenv
import logging
import sys
import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration


class Log:
    '''
    'debug'
    'info'
    'warning'
    'error'
    'critical'

    remeber me -- principle context propagation say

    '''

    def __init__(
        self,
        path_file: str = None,
        level: int = None,
        is_local_env: bool = False,
        event_level = None
    ) -> None:

        self.is_local_env = is_local_env
        self.filename = path_file
        self.LOGFORMAT = '%(asctime)s:%(levelname)s:%(name)s:%(message)s'
        self.DEFAULT_LEVEL = (
            'info'  # Your default level, usually set to warning or error for production
        )

        self.level = level
        self.event_level = event_level

        self.LEVELS = {
            'debug': DEBUG,
            'info': INFO,
            'warning': WARNING,
            'error': ERROR,
            'critical': CRITICAL,
        }

    # Iniciar logs
    def startlogging(self):

        basicConfig(
            level=self.LEVELS[self.level],
            format=self.LOGFORMAT,
            force=True,
            handlers=[
                logging.FileHandler(self.filename),
                logging.StreamHandler(sys.stdout)
            ]
        )

        if self.is_local_env is not True:
            # All of this is already happening by default!
            sentry_logging = LoggingIntegration(
                # Capture info and above as breadcrumbs
                level = self.LEVELS[self.level],
                # Send errors as events
                event_level = self.LEVELS[self.event_level],
            )
            sentry_sdk.init(
                dsn=getenv('SENTRY_DSN'),
                environment = getenv("NODE_ENV"),
                integrations=[
                    sentry_logging,
                ],
                # Set traces_sample_rate to 1.0 to capture 100%
                # of transactions for performance monitoring.
                # We recommend adjusting this value in production,
                traces_sample_rate=1.0,
            )

    def get_log(self, message):
        return getLogger(message)



'''
This class describes big chunks of text that may contain date strings
Each chunk includes of one of more tokens
Each token is build upon DATE_REGEX matches
'''


class DateFragment:
    def __init__(self):
        self.match_str = ''
        self.indices = (0, 0)
        self.captures = {}

    def __repr__(self):
        str_capt = ', '.join(['"{}": [{}]'.format(c, self.captures[c]) for c in self.captures])
        return '{} [{}, {}]\nCaptures: {}'.format(self.match_str, self.indices[0], self.indices[1], str_capt)

    def get_captures_count(self):
        return sum([len(self.captures[m]) for m in self.captures])



def divide_chunks(l: list = None, n: int = 100) -> list:

    # looping till length l
    for i in range(0, len(l), n):
        yield l[i:i + n]

def get_full_class_name(obj):
    """
    get full class name error 
    """
    module = obj.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return obj.__class__.__name__
    return module + '.' + obj.__class__.__name__



from datetime import datetime
from datetime import date
import numpy as np
import pandas as pd
from datetime import datetime,timedelta

from klp_commons.controllers.controller_redshift import ControllerRedShift
from klp_commons.controllers.controller_mongodb import ControllerMongoDB
from klp_commons.controllers.controller_postgresdb import ControllerPostgresDB

null = None
true = True
false = False

class Summary:
    """
    """

    # Variables de clase

    def __init__(self,from_date_str = None, num_days= 1):
        self.MongoDB = ControllerMongoDB()
        self.MongoDB.get_con_collect(collection_name='fact')

        self.num_days = num_days
        self.last_period = None
        self.from_date = from_date_str
        self.to_date = None
        self.set_date_range()
        
        
   
        
    def set_date_range(self):
        
        self.from_date = datetime.strptime(self.from_date , '%Y-%m-%d')
        self.to_date = self.from_date - timedelta(days=self.num_days)
        self.last_period = self.from_date + timedelta(days=self.num_days)
        print("self.from_date :", self.from_date )
        print("self.to_date :", self.to_date )
        print("self.last_period", self.last_period )
                        
        

        
    def agg_master(self):
        self.query_summary_category = [
            # { "$unwind": "$history-track" },
            {
            "$match": {
                    "$or":[
                        {"created_at": {'$gte':self.to_date,'$lt': self.from_date}},
                        {"created_at_homologation": {'$gte':self.to_date,'$lt': self.from_date}}
                        ],
                    "$and": [ { "history-track.validated": True}]
                    }
                        },
        {'$group': 
                 {'_id': {
                    "user_uid": "$user_uid",
                    "transaction_uid":"$_id",
                    "lang_code":"$lang_code",
                    "country_code":"$country_code",
                    "type_flow":"$type_flow",
                    "message_type_head":"$message_type",
                    "description":"$source_description",
                    "created_at_homologation":"$created_at_homologation",
                    "created_at_head": "$created_at" ,
                    "account_uid":"$account_number",
                    "bank_name":"$account_institution_name",
                    "account_type" :"$account_type",
                    "bank_type" :"NO_DISPONIBLE",

                    "message_type_history":"$history-track.message_type",
                    "message_type_old":"$history-track.type" ,

                    "model_nlp_version":"$history-track.model_nlp_version", # pre
                    "model_classification_version":"$history-track.model_classification_version",
                    "model_collective_version":"$history-track.model_collective_version",
                    "model_nickname_version":"$history-track.model_nickname_version",
                    "clean_description": "$history-track.clean_description",
                    "classified_date":"$history-track.classification_datetime",

                    "validated":"$history-track.validated", ##validated

                    "approved_asset":"$history-track.approved_asset",
                    "approved_business":"$history-track.approved_business",
                    "approved_category":"$history-track.approved_cat",
                    "approved_frequency":"$history-track.approved_frequency",
                    "approved_is_essential":"$history-track.approved_is_essential",
                    "approved_is_membership":"$history-track.approved_is_membership",
                    "approved_nickname":"$history-track.approved_nickname",
                    "approved_person":"$history-track.approved_person",
                    "approved_special_moment":"$history-track.approved_special_moment",
                    "approved_subcategory":"$history-track.details.approved_subcat",
                    "approved_is_loan_received":"$history-track.details.approved_is_loan_received",
                    "approved_is_loan_requested":"$history-track.details.approved_is_loan_requested",
                    "approved_income_source_uid":"$history-track.details.approved_income_source_uid",
                    
                    "category_custom_type":"$history-track.category_custom_type",
                    "subcategory_custom_type":"$history-track.subcategory_custom_type",
                     
                    "created_at":"$history-track.details.created_at",
                    "updated_at":"$history-track.details.updated_at",
                     
                     
                    "asset_uid_def_algo":"$history-track.details.asset_uid_def_algo",
                    "business_uid_def_algo":"$history-track.details.business_uid_def_algo",
                    "cat_uid_def_algo":"$history-track.details.cat_uid_def_algo",
                    "frequency_uid_def_algo":"$history-track.details.frequency_uid_def_algo",
                    "income_source_uid_def_algo":"$history-track.details.income_source_uid_def_algo",
                    "is_essential_def_algo":"$history-track.details.is_essential_def_algo",
                    "is_loan_received_def_algo":"$history-track.details.is_loan_received_def_algo",
                    "is_loan_requested_def_algo":"$history-track.details.is_loan_requested_def_algo",
                    "is_membership_def_algo":"$history-track.details.is_membership_def_algo",
                    "nickname_algo":"$history-track.details.nickname_algo",
                    "person_uid_def_algo":"$history-track.details.person_uid_def_algo",
                    "special_moment_uid_def_algo":"$history-track.details.special_moment_uid_def_algo",
                    "subcat_uid_def_algo":"$history-track.details.subcat_uid_def_algo",
                     

                    "asset_uid_def_usr":"$history-track.details.asset_uid_def_usr",
                    "asset_uid_def_usr":"$history-track.details.asset_uid_def_usr",
                    "business_uid_def_usr":"$history-track.details.business_uid_def_usr",
                    "cat_uid_def_usr":"$history-track.details.cat_uid_def_usr",
                    "frequency_uid_def_usr":"$history-track.details.frequency_uid_def_usr",
                    "income_source_uid_def_usr":"$history-track.details.income_source_uid_def_usr",
                    "is_essential_def_usr":"$history-track.details.is_essential_def_usr",
                    "is_loan_received_def_usr":"$history-track.details.is_loan_received_def_usr",
                    "is_loan_requested_def_usr":"$history-track.details.is_loan_requested_def_usr",
                    "is_membership_def_usr":"$history-track.details.is_membership_def_usr",
                    "nickname_usr":"$history-track.details.nickname_usr",
                    "nickname_user":"$history-track.details.nickname_user", #user
                    "person_uid_def_usr":"$history-track.details.person_uid_def_usr",
                    "special_moment_uid_def_usr":"$history-track.details.special_moment_uid_def_usr",
                    "subcat_uid_def_usr":"$history-track.details.subcat_uid_def_usr",
            
                    "asset_uid_def_usr_":"$history-track.details.info.asset_uid",
                     "business_uid_def_usr_":"$history-track.details.info.business_uid",
                     "created_at_def_usr_":"$history-track.details.info.created_at",
                     "frequency_uid_def_usr_":"$history-track.details.info.frequency_uid",
                     "income_source_uid_def_usr_":"$history-track.details.info.income_source_uid",
                     "is_essential_def_usr_":"$history-track.details.info.is_essential",
                     "is_loan_received_def_usr_":"$history-track.details.info.is_loan_received",
                     "is_loan_requested_def_usr_":"$history-track.details.info.is_loan_requested",
                     "is_membership_def_usr_":"$history-track.details.info.is_membership",
                     "nickname_usr_":"$history-track.details.info.name",
                     "person_uid_def_usr_":"$history-track.details.info.person_uid",
                     "special_moment_uid_def_usr_":"$history-track.details.info.special_moment_uid",

                     "asset_uid_def_usr_info":"$history-track.info.asset_uid",
                     "business_uid_def_usr_info":"$history-track.info.business_uid",
                     "created_at_def_usr_info":"$history-track.info.created_at",
                     "frequency_uid_def_usr_info":"$history-track.info.frequency_uid",
                     "income_source_uid_def_usr_info":"$history-track.info.income_source_uid",
                     "is_essential_def_usr_info":"$history-track.info.is_essential",
                     "is_loan_received_def_usr_info":"$history-track.info.is_loan_received",
                     "is_loan_requested_def_usr_info":"$history-track.info.is_loan_requested",
                     "is_membership_def_usr_info":"$history-track.info.is_membership",
                     "person_uid_def_usr_info":"$history-track.info.person_uid",
                     "special_moment_uid_def_usr_info":"$history-track.info.special_moment_uid",
                     
                    "asset_uid_def_collect":"$history-track.details.asset_uid_def_collect",
                    "business_uid_def_collect":"$history-track.details.business_uid_def_collect",
                    "cat_uid_def_collect":"$history-track.details.category_uid_def_collect",
                    "frequency_uid_def_collect":"$history-track.details.frequency_uid_def_collect",
                    "income_source_uid_def_collect":"$history-track.details.income_source_uid_def_collect",
                    "is_essential_def_collect":"$history-track.details.is_essential_def_collect",
                    "is_loan_received_def_collect":"$history-track.details.is_loan_received_def_collect",
                    "is_loan_requested_def_collect":"$history-track.details.is_loan_requested_def_collect",
                    "is_membership_def_collect":"$history-track.details.is_membership_def_collect",
                    "nickname_collect":"$history-track.details.nickname_collect",
                    "person_uid_def_collect":"$history-track.details.person_uid_def_collect",
                    "special_moment_uid_def_collect":"$history-track.details.special_moment_uid_def_collect",
                    "subcat_uid_def_collect":"$history-track.details.subcategory_uid_def_collect",
                      

                       },
                'count': {"$sum": 1}
                }

             
            }
            
        ]

        
       
        return pd.json_normalize(pd.DataFrame(list( self.MongoDB.con_db.aggregate(self.query_summary_category) ))['_id']) 

MongoDB = ControllerMongoDB()
PGDB = ControllerPostgresDB()
RedShift = ControllerRedShift()

RedShift.get_conection()

today = date.today()
front_date = today.strftime("%Y-%m-%d")

print(today)
print("front_date: " , front_date)

s = Summary(front_date,num_days = 30)
r = s.agg_master()



r['business_uid_def_usr'] = r['business_uid_def_usr'].apply(lambda x: x) + r['business_uid_def_usr_'].apply(lambda x: x) + r['business_uid_def_usr_info'].apply(lambda x: x)

del r['business_uid_def_usr_']
del r['business_uid_def_usr_info']


r['asset_uid_def_usr'] = r['asset_uid_def_usr'].apply(lambda x: x) + r['asset_uid_def_usr_'].apply(lambda x: x) + r['asset_uid_def_usr_info'].apply(lambda x: x)

del r['asset_uid_def_usr_info']
del r['asset_uid_def_usr_']

r['frequency_uid_def_usr'] = r['frequency_uid_def_usr'].apply(lambda x: x) + r['frequency_uid_def_usr_'].apply(lambda x: x) + r['frequency_uid_def_usr_info'].apply(lambda x: x)

del r['frequency_uid_def_usr_info']
del r['frequency_uid_def_usr_']

r['income_source_uid_def_usr'] = r['income_source_uid_def_usr'].apply(lambda x: x) + r['income_source_uid_def_usr_'].apply(lambda x: x) + r['income_source_uid_def_usr_info'].apply(lambda x: x)

del r['income_source_uid_def_usr_info']
del r['income_source_uid_def_usr_']


r['is_loan_received_def_usr'] = r['is_loan_received_def_usr'].apply(lambda x: x) + r['is_loan_received_def_usr_'].apply(lambda x: x) + r['is_loan_received_def_usr_info'].apply(lambda x: x)

del r['is_loan_received_def_usr_info']
del r['is_loan_received_def_usr_']

r['is_loan_requested_def_usr'] = r['is_loan_requested_def_usr'].apply(lambda x: x) + r['is_loan_requested_def_usr_'].apply(lambda x: x) + r['is_loan_requested_def_usr_info'].apply(lambda x: x)

del r['is_loan_requested_def_usr_info']
del r['is_loan_requested_def_usr_']


r['is_membership_def_usr'] = r['is_membership_def_usr'].apply(lambda x: x) + r['is_membership_def_usr_'].apply(lambda x: x) + r['is_membership_def_usr_info'].apply(lambda x: x)

del r['is_membership_def_usr_info']
del r['is_membership_def_usr_']


r['person_uid_def_usr'] = r['person_uid_def_usr'].apply(lambda x: x) + r['person_uid_def_usr_'].apply(lambda x: x) + r['person_uid_def_usr_info'].apply(lambda x: x)

del r['person_uid_def_usr_info']
del r['person_uid_def_usr_']


r['special_moment_uid_def_usr'] = r['special_moment_uid_def_usr'].apply(lambda x: x) + r['special_moment_uid_def_usr_'].apply(lambda x: x) + r['special_moment_uid_def_usr_info'].apply(lambda x: x)

del r['special_moment_uid_def_usr_info']
del r['special_moment_uid_def_usr_']


for col in sorted(r.columns.tolist()):
    r[col] = r.explode(col).reset_index()[col]

    
#r[cols].set_index(['nickname_def_usr','category_uid_def_algo']).apply(pd.Series.explode).reset_index()
# .sort_values(by=['created_at_head'], ascending = False)
print("shape: ", r.shape)

r.replace(np.nan, None, inplace=True)
r.replace(pd.NaT, None, inplace=True)
tmp = {'created_at_homologation': pd.NaT}


for idx ,item in enumerate(r.to_dict(orient='records')):
    print(idx, ": ",item['transaction_uid'])
    print("... ")
    if 'classification_datetime' in item:
        item['classified_at'] = item['classification_datetime']

    if not item['lang_code']:
        item['language_code'] = 'es-MX'
    else :
        item['language_code'] = 'es-MX'
        

    # Dict to necessary data
    transaction_dict = Transaction_().get()
    transaction_dict.update(item)


    transaction_dict['message_type'] = item['message_type_old']
    transaction_dict['flow_type'] = ''
    
    if not transaction_dict['clean_description']:
        transaction_dict['clean_description'] = ''


    if not transaction_dict['source_description']:
        transaction_dict['source_description'] = ''

    if not transaction_dict['model_classification_version']:
        transaction_dict['model_classification_version'] = ''

    if not transaction_dict['model_nickname_version']:
        transaction_dict['model_nickname_version'] = ''

    if not transaction_dict['model_nlp_version']:
        transaction_dict['model_nlp_version'] = ''

    if not transaction_dict['model_collective_version']:
        transaction_dict['model_collective_version'] = ''

        
    format_inter = '%Y-%m-%dT%H:%M:%S'
    
    if not transaction_dict['original_created_date'] or transaction_dict['original_created_date'] is pd.NaT:
        transaction_dict['original_created_date'] = datetime.strptime('2000-01-01T00:00:00',format_inter).replace(microsecond=0).isoformat()


    if not transaction_dict['created_at_homologation'] or transaction_dict['created_at_homologation'] is pd.NaT:
        transaction_dict['created_at_homologation'] = datetime.strptime('2000-01-01T00:00:00',format_inter).replace(microsecond=0).isoformat()

    if not item['created_at_head'] or item['created_at_head'] is pd.NaT:
        item['created_at_head'] = datetime.strptime('2000-01-01T00:00:00',format_inter).replace(microsecond=0).isoformat()

       
        
        
    if not transaction_dict['classified_at']:
        transaction_dict['classified_at'] = datetime.strptime('2000-01-01T00:00:00',format_inter).replace(microsecond=0).isoformat()

    if not transaction_dict['updated_at']:
        transaction_dict['updated_at'] = datetime.strptime('2000-01-01T00:00:00',format_inter).replace(microsecond=0).isoformat()
        
        
    if not transaction_dict['created_at']:
        transaction_dict['created_at'] = datetime.strptime('2000-01-01T00:00:00',format_inter).replace(microsecond=0).isoformat()
        
        
    if not  transaction_dict['message_type'] :
        transaction_dict['message_type'] = 'NO_FOUND'
    # print("transaction_dict['created_at']:::::::::::::", transaction_dict['created_at'])
    # if transaction_dict['created_at'] is pd.NaT:
    #    print("siiiiiii")
   

    keys = Transaction_().get().keys()

    rm_list = list()
    for key in transaction_dict.keys():
        if not key in keys:
            #print("No key : ", key)
            rm_list.append(key)

    for key in rm_list:
        del transaction_dict[key]

    # if message['type'] == 'report-expense-created':
    MongoDB.get_con_collect(collection_name = 'fact',db_name="klopp")
    
    
    user_dict = None
    # new user ?
    if not MongoDB.exist('user_uid',item['user_uid']):
        user_dict ={"user_uid":item['user_uid'],
                    "postcode":None
                    }

        # account_dict ={'account_uid': item['account']['account_number'],
        #                'bank_name':  item['account']['account_institution_name'],
        #                'account_type': item['account']['account_type'],
        #                'bank_type': item['account']['account_category']
        #                }
        RedShift.insert_dim_user(dict_values = user_dict)
        # self.RedShift.insert_dim_account(dict_values = account_dict)
        
        
        
    # print(user_dict)
    # print("****************************************\n")
    # print(transaction_dict)
    
    

    date_dict = RedShift.date_decomposition(str(item['created_at_head']).replace(" ","T").split(".")[0])
    table_name = 'dim_dates'
    column_name = 'transaction_date'

    if not RedShift.exists(table_name,column_name ,value = str(date_dict['transaction_date']) ):
        RedShift.insert_dim_dates(dict_values = date_dict)


    fact_transactions_dict = dict ()
    fact_transactions_dict['transaction_date'] = str(date_dict['transaction_date'])
    fact_transactions_dict['transaction_uid'] = item['transaction_uid']
    fact_transactions_dict['user_uid'] = item['user_uid']
    fact_transactions_dict['account_uid'] = ''

    #print("----------------------------------------------")
    #print("\n")
    # print(transaction_dict)

    # Validación o updated de una transacción 
    # poner la bandera de Categprización automática
    RedShift.insert_dim_transaction(dict_values = transaction_dict)

    RedShift.insert_fact_transactions(dict_values = fact_transactions_dict)



import googlemaps


def google_places(text: str = None) -> str:
    '''
    Petición a la API GOOLE PLACE para buscar el lugar almacenado en la cadena de texto de la varibale text


    '''

    response = googlemaps.places(query=text)
    if response['status'] == 'OK':
        results = response.get('results')[0]
        name = results['name']
    else:
        name = 'N/A'

    return name



from logging import getLogger
from fastapi.responses import JSONResponse
from os import getenv
from boto3 import resource
# User CRUD
from botocore.exceptions import ClientError
from boto3.dynamodb.conditions import Key, Attr
from klp_commons.log.structuredMessage import StructuredMessage
fmt = StructuredMessage
message = 'dynamoDBController'


class ControllerDynamoDB:
    """
    comment
    """
    def __init__(self):

        self.dyn_resource = None
        self.table_name_conection = None
        self.active_conection_table = None
        self.error = False
        self.aws_access_key_uid = None
        self.aws_secret_access_key = None
        self.region_name = None
        self.dyn_resource = None
        self.AWS_ACCESS_KEY_ID = getenv("AWS_ACCESS_KEY_ID")
        self.AWS_SECRET_ACCESS_KEY = getenv("AWS_SECRET_ACCESS_KEY")
        self.AWS_DEFAULT_REGION = getenv("AWS_DEFAULT_REGION")
        self.prefix = getenv("NODE_ENV") + '_'

        # module_name.TopicKafka_name
        self.logger = getLogger(message)

        # Init ()
        # MicroService_name.TopicKafka_name.module_name.MainClass_name
        self.logger.info("donde init method of ControllerDynamoDB ")
    
    def get_all_items_table(self):

        response = self.active_conection_table.scan(FilterExpression='attribute_not_exists(details.cat_uid_def_algo) and attribute_exists(info.income_source_uid)')
        data = response['Items']

        while 'LastEvaluatedKey' in response:
            response = self.active_conection_table.scan(ExclusiveStartKey=response['LastEvaluatedKey'])
            data.extend(response['Items'])

        return data

    def get_conection(self, table_name: str) -> None:

        self.logger.info("call get_conection() method of ControllerDynamoDB ")

        try:
            self.table_name_conection = self.prefix + table_name

            if self.dyn_resource is None:
                self.dyn_resource = resource(
                    "dynamodb",
                    aws_access_key_id=self.AWS_ACCESS_KEY_ID,
                    aws_secret_access_key=self.AWS_SECRET_ACCESS_KEY,
                    region_name=self.AWS_DEFAULT_REGION)

            self.active_conection_table = self.dyn_resource.Table(
                self.table_name_conection
            )
        except ClientError as err:
            self.logger.error(
                "Couldn't conected to table %s. Here's why: %s: %s",
                table_name,
                err.response["Error"]["Code"],
                err.response["Error"]["Message"],
            )
            self.logger.error(
                fmt("table_name ", json={'user_uid': table_name}))

        self.logger.info("done get_conection() method of ControllerDynamoDB ")

    def get_update_params(self, body, key_sorted, key_partition = 'user_uid'):
        """Given a dictionary we generate an update expression and a dict of values
        to update a dynamodb table.

        Params:
            body (dict): Parameters to use for formatting.

        Returns:
            update expression, dict of values.
        """
        keys_search = dict()
        update_expression = ["set "]
        update_values = dict()

        keys_search[key_partition] = body[key_partition]
        keys_search[key_sorted] = body[key_sorted]

        del body[key_partition]
        del body[key_sorted]
        
        for key, val in body.items():
            update_expression.append(f" {key} = :{key},")
            update_values[f":{key}"] = val

        return keys_search,"".join(update_expression)[:-1], update_values

    def update(self, body, key_sorted, key_partition = 'user_uid'):
        """
        
        
        """
        s, a, v = self.get_update_params(body, key_sorted, key_partition )
        
        response = self.active_conection_table.update_item(
            Key = s,
            UpdateExpression = a,
            ExpressionAttributeValues = dict(v)
            )
        return response

    def insert_hash_in_userTable(self, transaction: dict) -> dict:
        """
        transaction =  {
            "user_uid":"soyYai3q09Anhi1vzxJhgmqOMTLS5eXiKOWbTWsA",
            "description_embedding":"0101510102030101010101010101010101001",
            "data":
            {
                'subcat_uid_def_usr':"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz",
                'subcat_uid_def_algo':"U7CnAFNoegobBd4wvRu948iiqzCZ7glvIkB6PIxM",
                'subcat_uids_def_collect':{},
                'category_uid':"QsizWDPMOedPEzA5pvLaDEcxGqjHpInp1rOODZ9W"
            }
            }
        """

        try:
            self.r_query_dydb = self.active_conection_table.put_item(
                Item={
                    "user_uid": transaction["user_uid"],
                    "description_embedding": transaction["description_embedding"],
                    "language_code": transaction["language_code"],
                    "data": transaction["detail"],
                },
                ReturnConsumedCapacity="TOTAL"
                )

        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500
            )
            self.logger.error(fmt("transaction : ", json=transaction))

    def insert_batch_hash_in_userTable(self, items_list: list) -> dict:

        try:
            with self.active_conection_table.batch_writer(overwrite_by_pkeys=["user_uid", "hash"]) as writer:
                for item in items_list:
                    writer.put_item(Item=item)
            self.logger.info("Loaded data into table %s.",
                             self.active_conection_table.name)
        except ClientError:
            self.logger.exception(
                "Couldn't load data into table %s.", self.active_conection_table.name)
            Exception('error  insert_batch_hash_in_userTable DynamoDB!')

    def get_all_info_in_userTable(  self, 
                                partition_key_value: str
                                ) -> None:
        """
        partition_key_name: user_uid

        sort_key_name: hash resultaante de transformar la descripción (texto) de un atransacción
        (aka description_enbedding)
        """

        self.partition_key_name = "user_uid"
        try:

            self.r_query_dydb = self.active_conection_table.query(
                KeyConditionExpression=Key(self.partition_key_name).eq(
                    partition_key_value
                )
            )

        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500
            )
            self.logger.error(fmt("user iud ", json={
                'partition_key_value': partition_key_value}))

    def get_info_in_userTable(  self, 
                                partition_key_value: str,
                                sort_key_value: str
                            ) -> None:
        """
        partition_key_name: hash del user (user_ui)

        sort_key_name: hash resultaante de transformar la descripción (texto) de un atransacción
        (aka description_enbedding)

        Response :
        {'Items': [{'language_code': 'es-MX',
        'details': {'merch_name': 'Obrien-Miller',
            'subcat_uid_def_collect': '',
            'subcat_uid_def_algo': 'WIIxmmfsivKM4vBGbZCf3YcAAHFYDV48S8py15Np',
            'nickname': 'gimnasio',
            'cat_uid': 'csI1MhFMuv7Gp4URVyiCXWSIzSSXOzUN8MIWHVtP',
            'subcat_uid_def_usr': '',
            'stem': 'gimnasio taco'},
        'user_uid': '9ZM5VDfrDJ9NjE6nvYrbd04qvgdSgS04rQbvAylI',
        'description_embedding': '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000001000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000100000000000000000000000000000000000'}],
        'Count': 1,
        'ScannedCount': 1,
        'ResponseMetadata': {'RequestId': '0QIC45SH70CL8FAS1PJ3EHHK17VV4KQNSO5AEMVJF66Q9ASUAAJG',
        'HTTPStatusCode': 200,
        'HTTPHeaders': {'server': 'Server',
        'date': 'Wed, 10 Aug 2022 05:04:16 GMT',
        'content-type': 'application/x-amz-json-1.0',
        'content-length': '724',
        'connection': 'keep-alive',
        'x-amzn-requestid': '0QIC45SH70CL8FAS1PJ3EHHK17VV4KQNSO5AEMVJF66Q9ASUAAJG',
        'x-amz-crc32': '1412129221'},
        'RetryAttempts': 0}}
        """

        self.partition_key_name = "user_uid"
        self.sort_key_name = "description_embedding"
        try:

            self.r_query_dydb = self.active_conection_table.query(
                KeyConditionExpression=Key(self.partition_key_name).eq(
                    partition_key_value
                )
                & Key(self.sort_key_name).eq(sort_key_value),
            )

        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500
            )
            self.logger.error(fmt("user iud ", json={
                'partition_key_value': partition_key_value,
                "sort_key_value": sort_key_value}))

    def insert_popular_categories_in_multicategoryTable(
            self, transaction: dict) -> dict:
        """
        send in this format:

        transaction =  {
            "description_embedding":"0101510102030101010101010101010101001",
            "popular_subcategories": {1:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz",
            2:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz",
            3:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"

            },
            "categories": {1:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz",
            2:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz",
            3:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"

            }


            store format in AWS DynamoDB:

                        {
              "description_embedding": {
                "S": "0101510102030101010101010101010101001"
              },
              "categories": {
                "M": {
                  "1": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  },
                  "2": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  },
                  "3": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  }
                }
              },
              "popular_subcategories": {
                "M": {
                  "1": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  },
                  "2": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  },
                  "3": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  }
                }
              }
            }

        """
        self.logger.info(
            "call insert_popular_categories_for_transaction() method of ControllerDynamoDB ")

        try:
            self.r_query_dydb = self.active_conection_table.put_item(
                Item={
                    "description_embedding": transaction["description_embedding"],
                    "popular_subcategories": transaction["popular_subcategories"],
                    "categories": transaction["categories"],
                })
        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500)
            self.logger.error(fmt("transaction :", json=transaction))

        self.logger.info(
            "done insert_popular_categories_in_multicategoryTable() method of ControllerDynamoDB ")

    def insert_bacth_in_multicategoryTable(self, items_list: list) -> dict:
        """
        send in this format:

        transaction =  {
            "description_embedding":"0101510102030101010101010101010101001",
            "popular_subcategories": {1:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz",
            2:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz",
            3:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"

            },
            "categories": {1:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz",
            2:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz",
            3:"JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"}

            store format in AWS DynamoDB:

                        {
              "description_embedding": {
                "S": "0101510102030101010101010101010101001"
              },
              "categories": {
                "M": {
                  "1": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  },
                  "2": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  },
                  "3": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  }
                }
              },
              "popular_subcategories": {
                "M": {
                  "1": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  },
                  "2": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  },
                  "3": {
                    "S": "JxdCwO7AA5iuQUeuc1M2o2XQZJsIq0iaucoai9Kz"
                  }
                }
              }
            }

        """
        self.logger.info(
            "call insert_popular_categories_for_transaction() method of ControllerDynamoDB ")


        try:
            with self.active_conection_table.batch_writer(overwrite_by_pkeys=["user_uid", "description_embedding"]) as writer:
                for item in items_list:
                    writer.put_item(Item=item)

        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500)
            self.logger.error(fmt("transaction :", json=items_list))


        self.logger.info(
            "donde insert_bacth_in_multicategoryTable() method of ControllerDynamoDB ")

    def get_popular_categories_in_multicategoryTable(
            self, partition_key_value: str) -> None:
        """
        partiton_key_name: hash de la descripción de un atransacción (description_enbedding)
        """

        self.logger.info(
            "call get_popular_categories_in_transactionTable() method of ControllerDynamoDB ")

        self.partition_key_name = "description_embedding"
        self.column = "popular_subcategories"
        try:

            self.r_query_dydb = self.active_conection_table.query(
                KeyConditionExpression=Key(self.partition_key_name).eq(
                    partition_key_value
                ),
                ProjectionExpression=self.column,
            )

        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500
            )
            self.logger.error(
                fmt("user iud ", json={'partition_key_value': partition_key_value}))

        self.logger.info(
            "done get_popular_categories_in_multicategoryTable() method of ControllerDynamoDB ")

    def insert_info_in_transactionTable(self, transaction: dict) -> dict:
        """
        transaction =  {
            "user_uid":"soyYai3q09Anhi1vzxJhgmqOMTLS5eXiKOWbTWsA",
            "transaction_uid":"0101510102030101010101010101010101001",
            "hash": "0101510102030101010101010101010101001"
                        }
        """
        self.logger.info(
            "call insert_hash_in_userTable_user() method of ControllerDynamoDB ")
        try:
            self.r_query_dydb = self.active_conection_table.put_item(
                Item={
                    "user_uid": transaction["user_uid"],
                    "transaction_uid": transaction["transaction_uid"],
                    "hash": transaction["hash"],
                    "nickname": transaction["nickname"],
                },
                ReturnConsumedCapacity="TOTAL"
            )

        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500
            )
            self.logger.error(fmt("user iud ", json=transaction))


        self.logger.info(
            "done insert_info_in_transactionTable method of ControllerDynamoDB ")

    def insert_bacth_in_transactionTable(self, items_list: list) -> dict:
        """
        transaction =  {
            "user_uid":"soyYai3q09Anhi1vzxJhgmqOMTLS5eXiKOWbTWsA",
            "transaction_uid":"0101510102030101010101010101010101001",
            "hash": "0101510102030101010101010101010101001"
            "nickname": ""
                        }
        """
        self.logger.info(
            "call insert_bacth_in_transactionTable method of ControllerDynamoDB ")
        try:
            with self.active_conection_table.batch_writer(overwrite_by_pkeys=["user_uid", "transaction_uid"]) as writer:
                for item in items_list:
                    writer.put_item(Item=item)

        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500
            )
            self.logger.error(fmt("user iud ", json=items_list))


        self.logger.info(
            "done insert_bacth_in_transactionTable() method of ControllerDynamoDB ")

    def get_all_hashes_in_transactionTable(
        self,
        partition_key_value: str,
        hash:str
    ) -> None:
        """
        : hash del user (user_uid)

        """
        self.logger.info(
            "call get_hash_transaction_table() method of ControllerDynamoDB ")

        self.partition_key_name = "user_uid"

        try:

            self.r_query_dydb = self.active_conection_table.query(
                KeyConditionExpression=Key(self.partition_key_name).eq(partition_key_value),
                FilterExpression=Attr('hash').eq(hash)

                )

        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500
            )
            self.logger.error(fmt("user iud ", json={
                'partition_key_value': partition_key_value,'content':self.r_query_dydb}))

        self.logger.info(
            "done get_all_hashes_in_transactionTable method of ControllerDynamoDB ")

    def get_hash_in_transactionTable(
        self,
        partition_key_value: str,
        sort_key_value: str,
    ) -> None:
        """
        : hash del user (user_uid)

        : hash resultaante de transformar la descripción (texto) de un atransacción
        (aka description_enbedding)

        """
        self.logger.info(
            "call get_hash_transaction_table() method of ControllerDynamoDB ")

        self.partition_key_name = "user_uid"
        self.sort_key_name = "transaction_uid"

        try:

            self.r_query_dydb = self.active_conection_table.query(
                KeyConditionExpression=Key(self.partition_key_name).eq(
                    partition_key_value
                )
                & Key(self.sort_key_name).eq(sort_key_value),
                # ProjectionExpression = self.hash
            )

        except ClientError as e:
            self.r_query_dydb = JSONResponse(
                content=e.response["Error"], status_code=500
            )
            self.logger.error(fmt("user iud ", json={
                'partition_key_value': partition_key_value, 'sort_key_value': sort_key_value}))

        self.logger.info(
            "done get_hash_in_transactionTable method of ControllerDynamoDB ")

    def created_info_in_userTable(self, info: dict = None):
        """
        Updates user table.
        """
        
        self.logger.info(
            "call created_info_in_userTable() method of ControllerDynamoDB ")

        try:
            self.r_query_dydb = self.active_conection_table.update_item(
                Key={
                    "user_uid": info["user_uid"],
                    "description_embedding": info["description_embedding"],
                },
                UpdateExpression =  "set validated=:a,\
                                        approved_nickname=:b,\
                                        approved_cat=:c,  \
                                        approved_subcat=:d, \
                                        #det.subcat_uid_def_usr=:l,\
                                        #det.subcat_uid_def_algo=:m,\
                                        #det.nickname_algo=:n,\
                                        #det.nickname_usr=:o,\
                                        #det.cat_uid_def_algo=:p,\
                                        #det.cat_uid_def_usr=:q,\
                                        #det.updated_at=:r\
                                        ",
                                        
                ExpressionAttributeNames={
                    "#det": "details"},
                ExpressionAttributeValues={

                    ":a": info['validated'],
                    ":b": info['approved_nickname'],
                    ":c": info['approved_cat'],
                    ":d": info['approved_subcat'],

                    ":l": info['details']['subcat_uid_def_usr'],
                    ":m": info['details']['subcat_uid_def_algo'],
                    ":n": info['details']['nickname_algo'],
                    ":o": info['details']['nickname_usr'],
                    ":p": info['details']['cat_uid_def_algo'],
                    ":q": info['details']['cat_uid_def_usr'],
                    ":r": info['details']["updated_at"],
                },
                ReturnValues="UPDATED_NEW",
            )

        except ClientError as e:
            if e.response["Error"]["Code"] == "ValidationException":
                # Creating new top level attribute `info` (with nested props)
                # if the previous query failed
                self.r_query_dydb = self.active_conection_table.update_item(
                    Key={
                        "user_uid": info["user_uid"],
                        "description_embedding": info["description_embedding"],
                    },
                    UpdateExpression="set validated=:a,\
                                        approved_nickname=:b,\
                                        approved_cat=:c,  \
                                        approved_subcat=:d, \
                                        #det = :attrValue",
                    ExpressionAttributeNames={
                        "#det": "details"},
                    ExpressionAttributeValues={
                        ":a": info['validated'],
                        ":b": info['approved_nickname'],
                        ":c": info['approved_cat'],
                        ":d": info['approved_subcat'],
                            ":attrValue": {
                                "subcat_uid_def_usr": info['details']["subcat_uid_def_usr"],
                                "subcat_uid_def_algo": info['details']["subcat_uid_def_algo"],
                                "nickname_algo": info['details']["nickname_algo"],
                                "nickname_usr": info['details']["nickname_usr"],
                                "cat_uid_def_algo": info['details']["cat_uid_def_algo"],
                                "cat_uid_def_usr": info['details']["cat_uid_def_usr"],
                                "updated_at": info['details']["updated_at"]                                        }
                        },
                    ReturnValues="UPDATED_NEW",
                )
            else:
                self.logger.exception(
                    fmt("message created_info_in_userTable", json=info))

        self.logger.info(
            "done created_info_in_userTable() method of ControllerDynamoDB")

    def update_info_in_userTable(self, info: dict = None):
        """
        Updates user table.

        """

        self.logger.info(
            "call update_info_in_userTable() method of ControllerDynamoDB ")

        try:
            self.r_query_dydb = self.active_conection_table.update_item(
                Key={
                    "user_uid": info["user_uid"],
                    "description_embedding": info["description_embedding"],
                },
                UpdateExpression="set #det.subcat_uid_def_usr=:l, \
                                        #det.cat_uid_def_usr=:m,\
                                        #det.nickname_usr=:n,\
                                        #det.updated_at=:o",
                ExpressionAttributeNames={
                    "#det": "details"},
                ExpressionAttributeValues={
                    ":l": info['details']['subcat_uid_def_usr'],
                    ":m": info['details']['cat_uid_def_usr'],
                    ":n": info['details']['nickname_usr'],
                    ":o": info['details']['updated_at'],
                },
                ReturnValues="UPDATED_NEW",
            )

        except ClientError as e:
            if e.response["Error"]["Code"] == "ValidationException":
                # Creating new top level attribute `info` (with nested props)
                # if the previous query failed
                self.r_query_dydb = self.active_conection_table.update_item(
                    Key={
                        "user_uid": info["user_uid"],
                        "description_embedding": info["description_embedding"],
                    },
                    UpdateExpression="set #det = :attrValue",
                    ExpressionAttributeNames={
                        "#det": "details"},
                    ExpressionAttributeValues={
                        ":attrValue": {
                            "subcat_uid_def_usr": info['details']["subcat_uid_def_usr"],
                            "cat_uid_def_usr": info['details']["cat_uid_def_usr"],
                            "nickname_usr": info['details']["nickname_usr"],
                            "updated_at": info['details']["updated_at"],
                                    }
                        },
                    ReturnValues="UPDATED_NEW",
                )
            else:
                self.logger.exception(
                    fmt("message update_info_in_userTable", json=info))


        self.logger.info(
            "done update_info_in_userTable() method of ControllerDynamoDB")

    def delete_register_userTable(self, user_uid, description_embedding):
        """ 

        """

        self.logger.info(
            "call delete_register_userTable() method of ControllerDynamoDB")
        try:
            self.r_query_dydb = self.active_conection_table.delete_item(
                Key={
                    "user_uid": user_uid,
                    "description_embedding": description_embedding,
                }
            )

        except ClientError as err:
            self.logger.error(
                "Couldn't delete register for User table %s. Here's why: %s: %s",
                user_uid,
                err.response["Error"]["Code"],
                err.response["Error"]["Message"],
            )
            self.logger.exception(
                fmt("user iud ", json={'user_uid': user_uid}))

        self.logger.info(
            "done delete_register_userTable() method of ControllerDynamoDB")

    def delete_register_transactionTable(self, user_uid, transaction_uid):
        """

        """

        self.logger.info(
            "call delete_register_transactionTable() method of ControllerDynamoDB")

        try:
            self.r_query_dydb = self.active_conection_table.delete_item(
                Key={"user_uid": user_uid, "transaction_uid": transaction_uid})

        except ClientError as err:
            self.logger.error(
                "Couldn't delete register for Transaction table %s. Here's why: %s: %s",
                user_uid,
                err.response["Error"]["Code"],
                err.response["Error"]["Message"],
            )
            self.logger.error(
                fmt("user iud ", json={'user_uid': user_uid}))

        self.logger.info(
            "donde delete_register_transactionTable() method of ControllerDynamoDB")

    def get_batch_info_in_userTable(self, items) -> None:
        """
        """

        try:
            if self.dyn_resource is None:
                self.dyn_resource = resource(
                    "dynamodb",
                    aws_access_key_id=self.AWS_ACCESS_KEY_ID,
                    aws_secret_access_key=self.AWS_SECRET_ACCESS_KEY,
                    region_name=self.AWS_DEFAULT_REGION)

            self.result_query_dynamodb = self.dyn_resource.batch_get_item(
                RequestItems=items)

        except ClientError as e:
            self.result_query_dynamodb = JSONResponse(
                content=e.response["Error"], status_code=500
            )
            self.logger.error(fmt("items ", json=items))
   
    def get_batch_info_in_multicategoryTable(self, items) -> None:
        """

        """
        try:
            if self.dyn_resource is None:
                self.dyn_resource = resource(
                    "dynamodb",
                    aws_access_key_id=self.AWS_ACCESS_KEY_ID,
                    aws_secret_access_key=self.AWS_SECRET_ACCESS_KEY,
                    region_name=self.AWS_DEFAULT_REGION)

            self.result_query_dynamodb = self.dyn_resource.batch_get_item(
                RequestItems=items)

        except ClientError as e:
            self.result_query_dynamodb = JSONResponse(
                content=e.response["Error"], status_code=500)
            self.logger.error(fmt("items ", json=items))

    def get_batch_info_in_collectiveTable(self, items) -> None:
        """
        # ["subcategory_uid"]
        # ["category_uid"]
        # ["collective_category_uid"]
        # ["collective_subcategory_uid"]
        # ["merchant_name"]
        # ['merch_uid']
        # ["description"]
        # ["collective_description"]
        # ['nickname']

        """
        try:
            if self.dyn_resource is None:
                self.dyn_resource = resource(
                    "dynamodb",
                    aws_access_key_id=self.AWS_ACCESS_KEY_ID,
                    aws_secret_access_key=self.AWS_SECRET_ACCESS_KEY,
                    region_name=self.AWS_DEFAULT_REGION)

            self.result_query_dynamodb = self.dyn_resource.batch_get_item(
                RequestItems=items)

        except ClientError as e:
            self.result_query_dynamodb = JSONResponse(
                content=e.response["Error"], status_code=500)
            self.logger.error(fmt("items ", json=items))

    def insert_bacth_hash_in_collectiveTable(self, items_list: list) -> dict:

        try:
            with self.active_conection_table.batch_writer(overwrite_by_pkeys=["user_uid", "description_embedding"]) as writer:
                for item in items_list:
                    writer.put_item(Item=item)
            self.logger.info("Loaded data into table %s.",
                            self.active_conection_table.name)
        except ClientError:
            self.logger.error(
                "Couldn't load data into table %s.", self.active_conection_table.name)

